/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#ifndef TCP_IPV4_SEGMENT_H
#define TCP_IPV4_SEGMENT_H

#include "ip/ipv4header.h"
#include "tcp/tcpheader.h"
#include "tcp/tcpsegment.h"

//-----------------------------------------------------------------------------
// TcpIpv4Segment
//-----------------------------------------------------------------------------

/**
 * @brief This class represents the TCP segment over IP version 4.
 *
 * The TCP Segment over IPv4 Packet format is:
 *
 * @code
 * 0               8               16                             31
 * +-------+-------+---------------+------------------------------+      ---
 * |       |       |               |                              |       ^
 * |version|header |    type of    |    total length in bytes     |       |
 * |  (4)  | length|    service    |                              |       |
 * +-------+-------+---------------+-+-+-+------------------------+       |
 * |                               | | | |                        |       |
 * |        identification         |0|D|M|    fragment offset     |       |
 * |                               | |F|F|                        |       |
 * +---------------+---------------+-+-+-+------------------------+       |
 * |               |               |                              |       |
 * | time to live  |   protocol    |       header checksum        |  IPv4 Header
 * |               |               |                              |   20 bytes
 * +---------------+---------------+------------------------------+       |
 * |                                                              |       |
 * |                      source IPv4 address                     |       |
 * |                                                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |       |
 * |                   destination IPv4 address                   |       |
 * |                                                              |       v
 * +-------------------------------+------------------------------+      ---
 * |                               |                              |       ^
 * |         source port           |     destination port         |       |
 * |                               |                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |       |
 * |                        sequence number                       |       |
 * |                                                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |  TCP Header
 * |              acknowledgment number (if ACK set)              |   20 bytes
 * |                                                              |       |
 * +-------+-------+-+-+-+-+-+-+-+-+------------------------------+       |
 * | data  | reser-|C|E|U|A|P|R|S|F|                              |       |
 * |offset | ved   |W|C|R|C|S|S|Y|I|        window size           |       |
 * |       |       |R|E|G|K|H|T|N|N|                              |       |
 * +---------------+-+-+-+-+-+-+-+-+------------------------------+       |
 * |                               |                              |       |
 * |          checksum             | urgent pointer (if URG set)  |       |
 * |                               |                              |       v
 * +-----------------------------------------------+--------------+      ---
 * @endcode
 */
class TcpIpv4Segment : public TcpSegment
{
public:
    TcpIpv4Segment();
    TcpIpv4Segment( const TcpHeader &tcp_header );
    virtual ~TcpIpv4Segment();

    Ipv4Header get_ip_header() const;
    TcpHeader get_tcp_header() const;

    virtual bool match_rst_reply(
            const boost::asio::ip::address &source_address
    ) const;

    virtual void print_rst_reply(
            const boost::posix_time::ptime &time_segment_sent
    ) const;

    virtual bool match_syn_ack_reply(
            const boost::asio::ip::address &source_address
    ) const;

    virtual void print_syn_ack_reply(
            const boost::posix_time::ptime &time_packet_sent
    ) const;

    virtual bool read( std::istream &is );
    virtual bool write( std::ostream &os ) const;

    friend std::istream& operator>>(
            std::istream &is,
            TcpIpv4Segment &segment
    );
    friend std::ostream& operator<<(
            std::ostream &os,
            const TcpIpv4Segment &segment
    );

private:
    /// The IP header.
    Ipv4Header IpHeader;
    /// The TCP header.
    TcpHeader TcpHdr;

};

#endif // TCP_IPV4_SEGMENT_H
